#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "liblihata/dom.h"
#include "liblihata/tree.h"
#include "liblihata/hash_str.h"
#include "lht_marshaller.h"

typedef enum {
	FT_INT,
	FT_DOUBLE,
	FT_FLOAT,
	FT_CSTR,
	FT_STRUCT,
	FT_STRUCT_PTR,
} ftype_t;


static int make_field_cache(lht_node_t *desc)
{
	lht_dom_iterator_t it;
	lht_node_t *f, *fields, *cache_node;
	int alloced = 0, used = 0, numfields;
	char *cache = NULL;

#define grow(n) \
	do { \
		if (used+n-1 >= alloced) { \
			alloced += 32+n; \
			cache = realloc(cache, alloced); \
		} \
	} while(0)

#define append(c) \
	do { \
		grow(1); \
		cache[used] = c; \
		used++; \
	} while(0)

#define append_int(i) \
	do { \
		int _n_, _i_ = i; \
		grow(5); \
		for(_n_ = 0; _n_ < 5; _n_++) { \
			int _b_ = _i_ & ((1 << 6) -1); \
			_b_ += '0'; \
			if (_b_ >= '\\') \
				_b_++; \
			cache[used] = _b_; \
			used++; \
			_i_ >>= 6; \
		} \
	} while(0)

#define append_str(str) \
	do { \
			int _l_ = strlen(str); \
			char *_s_; \
			append_int(_l_); \
			for(_s_ = str; *_s_ != '\0'; _s_++) \
				append(*_s_); \
	} while(0)

	fields = lht_dom_hash_get(desc, "fields");
	if (fields == NULL) {
		fprintf(stderr, "no fields in type desc\n");
		return -1;
	}

 /* placeholder for number of fields */
	grow(5); used = 5;
	for(numfields = 0, f = lht_dom_first(&it, fields); f != NULL; f = lht_dom_next(&it), numfields++) {
		if (strcmp(f->name, "int") == 0)          { append('a' + FT_INT); }
		else if (strcmp(f->name, "double") == 0)  { append('a' + FT_DOUBLE); }
		else if (strcmp(f->name, "float") == 0)   { append('a' + FT_FLOAT); }
		else if (strcmp(f->name, "cstr") == 0)    { append('a' + FT_CSTR); }
		else if (*f->name == '*') {
			append('a' + FT_STRUCT_PTR);
			append_str(f->name+1);
		}
		else {
			fprintf(stderr, "WHAT TYPE?! '%s'\n", f->name);
			return -1;
		}
		append_str(f->data.text.value);
	}
	append('\0');
	used = 0;
	append_int(numfields);
	cache_node = lht_dom_node_alloc(LHT_TEXT, "fields_cache");
	cache_node->data.text.value = cache;
	lht_dom_hash_put(desc, cache_node);
	return 0;
}


#define printf_type(_format_, _type_) \
do { \
	sprintf(val, _format_, *(_type_ *)native); \
	native += sizeof(_type_); \
} while(0) \

#define get_field_cache \
	fields = lht_dom_hash_get(desc, "fields_cache"); \
	if (fields == NULL) { \
		make_field_cache(desc); \
		fields = lht_dom_hash_get(desc, "fields_cache"); \
		printf("field_cache='%s'\n", fields->data.text.value); \
	}


lht_node_t *lht_marshaller_b2l(lht_node_t *desc, lht_doc_t *doc, void *native_)
{
	lht_node_t *ret, *fields, *f;
	lht_dom_iterator_t it;
	char *native = native_;

	get_field_cache;

	fields = lht_dom_hash_get(desc, "fields");
	if (fields == NULL)
		return NULL;

	ret = lht_dom_node_alloc(LHT_HASH, desc->name);

	for(f = lht_dom_first(&it, fields); f != NULL; f = lht_dom_next(&it)) {
		char val_[256], *val = val_;
		lht_node_t *n;

#warning TODO: use sphash here
		if (strcmp(f->name, "int") == 0) printf_type("%d", int);
		else if (strcmp(f->name, "double") == 0) printf_type("%.16f", double);
		else if (strcmp(f->name, "float") == 0) printf_type("%f", float);
		else if (strcmp(f->name, "cstr") == 0) { val = *(char **)native; native += sizeof(char *); }
		else goto err;
		n = lht_dom_node_alloc(LHT_TEXT, f->data.text.value);
		if (n == NULL)
			goto err;
		n->data.text.value = strdup(val);
		if (lht_dom_hash_put(ret, n) != LHTE_SUCCESS)
			goto err;
	}
	return ret;

	err:;
	lht_dom_node_free(ret);
	return NULL;
}

#define scanf_type(_format_, _type_) \
do { \
	sscanf(v->data.text.value, _format_, (_type_ *)native); \
	native += sizeof(_type_); \
} while(0) \

int lht_marshaller_l2b(lht_node_t *desc, void *native_, lht_node_t *node)
{
	lht_node_t *fields, *f;
	lht_dom_iterator_t it;
	char *native = native_;

	get_field_cache;


	fields = lht_dom_hash_get(desc, "fields");
	if (fields == NULL)
		return -1;

	for(f = lht_dom_first(&it, fields); f != NULL; f = lht_dom_next(&it)) {
		lht_node_t *v;

		v = lht_dom_hash_get(node, f->data.text.value);
		if (v == NULL) {
			fprintf(stderr, "MISSING FIELD?! '%s'\n", f->data.text.value);
			return -1;
		}

#warning TODO: use sphash here
		if (strcmp(f->name, "int") == 0) scanf_type("%d", int);
		else if (strcmp(f->name, "double") == 0) scanf_type("%lf", double);
		else if (strcmp(f->name, "float") == 0) scanf_type("%f", float);
		else if (strcmp(f->name, "cstr") == 0) { *(char **)native = strdup(v->data.text.value); native += sizeof(char *); }
		else if (*f->name == '*') { native += sizeof(void *); }
		else {
			fprintf(stderr, "WHAT TYPE?! '%s'\n", f->name);
			return -1;
		}
	}
	return 0;
}

int lht_marshaller_size(lht_node_t *desc)
{
	lht_node_t *fields, *f;
	lht_dom_iterator_t it;
	int size = 0;

	get_field_cache;

	fields = lht_dom_hash_get(desc, "fields");
	if (fields == NULL)
		return -1;

	for(f = lht_dom_first(&it, fields); f != NULL; f = lht_dom_next(&it)) {
#warning TODO: use sphash here
		if (strcmp(f->name, "int") == 0) size += sizeof(int);
		else if (strcmp(f->name, "double") == 0) size += sizeof(double);
		else if (strcmp(f->name, "float") == 0) size += sizeof(float);
		else if (strcmp(f->name, "cstr") == 0) size += sizeof(char *);
		else if (*f->name == '*') size += sizeof(void *);
		else {
			fprintf(stderr, "WHAT TYPE?! '%s'\n", f->name);
			return -1;
		}
	}
	return size;
}

struct lht_marshaller_ctx_s {
	lht_node_t *desc_tree;
	htsp_t *n2p;
};

lht_marshaller_ctx_t *lht_marshaller_ctx_init(lht_node_t *desc_tree)
{
	lht_marshaller_ctx_t *ctx;

	ctx = malloc(sizeof(lht_marshaller_ctx_t));
	ctx->desc_tree = desc_tree;
	ctx->n2p = htsp_alloc(lht_str_keyhash, lht_str_keyeq);
	return ctx;
}

void *lht_marshaller_ctx_load(lht_marshaller_ctx_t *ctx, lht_node_t *root, const char *retname)
{
	lht_dom_iterator_t it;
	lht_node_t *n;

	for(n = lht_dom_first(&it, root); n != NULL; n = lht_dom_next(&it)) {
		int size;
		void *native;
		char *desc_name;
		lht_node_t *desc;

		if (n->type != LHT_HASH) {
			fprintf(stderr, "data node is not a hash\n");
			return NULL;
		}

		desc = lht_dom_hash_get(n, "");
		if (desc == NULL) {
			fprintf(stderr, "data node does not have a type field\n");
			return NULL;
		}
		if (desc->type != LHT_TEXT) {
			fprintf(stderr, "type field is not text\n");
			return NULL;
		}

		desc_name = desc->data.text.value;

		desc = lht_dom_hash_get(ctx->desc_tree, desc_name);
		if (desc == NULL) {
			fprintf(stderr, "can't file description called '%s'\n", desc_name);
			return NULL;
		}


		size = lht_marshaller_size(desc);
		native = calloc(1, size);

printf("size=%d native=%p\n", size, native);

		if (lht_marshaller_l2b(desc, native, n) != 0) {
			fprintf(stderr, "failed to marshal data node\n");
			return NULL;
		}

		htsp_set(ctx->n2p, n->name, native);
	}

	return htsp_get(ctx->n2p, (char *)retname);
}

